{#-
    This macro replaces human readable strings with its RC4 encrypted string
    and decrypts it in place, so it will be opaque to anti malware scanners
-#}
{%- macro obfuscate(msg) -%}
({{symbol_name("Decrypt-String")}} "{{msg|rc4encrypt}}")
{%- endmacro -%}
