<div class="toast ml-auto {{show}} mt-3" {{hidden}}
   role="alert"
   aria-live="assertive"
   aria-atomic="true"
   data-delay="10000">
    <div class="toast-header">
      <strong class="mr-auto toast-title text-{{message['category']}}">
        {{message['title']}}
      </strong>
      <small class="toast-subtitle">{{message['subtitle']}}</small>
      <button type="button" class="ml-2 mb-1 close" data-dismiss="toast" aria-label="Close">
        <span aria-hidden="true">&times;</span>
      </button>
    </div>
    <div class="toast-body">{{message['body']}}</div>
</div>
