$Global:M365DSCEmojiAccessDenied = [char]::ConvertFromUtf32(0x1F6AB)
$Global:M365DSCEmojiAlert = [char]::ConvertFromUtf32(0x1F514)
$Global:M365DSCEmojiAnnouncement = [char]::ConvertFromUtf32(0x1F4E2)
$Global:M365DSCEmojiClipboard = [char]::ConvertFromUtf32(0x1F4CB)
$Global:M365DSCEmojiCloud = [char]::ConvertFromUtf32(0x2601)
$Global:M365DSCEmojiEnveloppe = [char]::ConvertFromUtf32(0x2709)
$Global:M365DSCEmojiFolder = [char]::ConvertFromUtf32(0x1F4C1)
$Global:M365DSCEmojiGreenCheckmark = [char]::ConvertFromUtf32(0x2705)
$Global:M365DSCEmojiHourglass = [char]::ConvertFromUtf32(0x231B)
$Global:M365DSCEmojiInformation = [char]::ConvertFromUtf32(0x2139)
$Global:M365DSCEmojiLightbulb = [char]::ConvertFromUtf32(0x1F4A1)
$Global:M365DSCEmojiLock = [char]::ConvertFromUtf32(0x1F512)
$Global:M365DSCEmojiPackage = [char]::ConvertFromUtf32(0x1F4E6)
$Global:M365DSCEmojiPencil = [char]::ConvertFromUtf32(0x270F)
$Global:M365DSCEmojiPlanet = [char]::ConvertFromUtf32(0x1F30F)
$Global:M365DSCEmojiPolice = [char]::ConvertFromUtf32(0x1F46E)
$Global:M365DSCEmojiRedX = [char]::ConvertFromUtf32(0x274C)
$Global:M365DSCEmojiReport = [char]::ConvertFromUtf32(0x1F4CA)
$Global:M365DSCEmojiRocket = [char]::ConvertFromUtf32(0x1F680)
$Global:M365DSCEmojiScript = [char]::ConvertFromUtf32(0x1F4DC)
$Global:M365DSCEmojiSmallBlueBullet = [char]::ConvertFromUtf32(0x1F539)
$Global:M365DSCEmojiSnowflake = [char]::ConvertFromUtf32(0x2744)
$Global:M365DSCEmojiSpeech = [char]::ConvertFromUtf32(0x1F4AC)
$Global:M365DSCEmojiTarget = [char]::ConvertFromUtf32(0x1F3AF)
$Global:M365DSCEmojiWarningSign = [char]::ConvertFromUtf32(0x26A0)
$Global:M365DSCEmojiWatch = [char]::ConvertFromUtf32(0x231A)
$Global:M365DSCEmojiWhale = [char]::ConvertFromUtf32(0x1F40B)
$Global:M365DSCEmojiWhiteCircle = [char]::ConvertFromUtf32(0x26AA)
$Global:M365DSCEmojiYellowCircle = [char]::ConvertFromUtf32(0x1F7E1)
