{% set hub_main = "This is the Hub. Most of its functionality lies in its
PowerShell script. Paste the command below in a PowerShell/Bash command/CMD
and follow the instructions. You will be able to easily transfer and execute
more PowerShell modules and binary code in-memory (compressed and
encrypted). The endpoint protection will most likely not care. Note that you
need to place or symlink the modules in the <code>module</code> directory
first. The stuff in <code>$XDG_DATA_HOME/powerhub/profile.ps1</code> will
be executed automatically. Have fun!" %}

{% set modules_main = "These are your Modules. Here you can view available
modules and download more." %}

{% set modules = "Click on 'clone' to perform a shallow clone of the
selected repository. Alternatively, you can specify a URL which will be
cloned if it ends in '.git', or downloaded if it ends in '.ps1' or '.exe'."
%}

{% set hub_no_modules = "You probably didn't put anything in the
<code>modules</code> directory located at
<code>$XDG_DATA_HOME/powerhub</code>. If you already have those modules on
your hard drive, it makes sense to symlink them. But if you don't, or if
you are really lazy, you can download some select offensive security tools
by clicking on \"Need more?\"." %}

{% set fileexchange_main = "This is the File Exchange. Here you can upload
and download files via HTTP. Think Droopy. The files are in your workspace
directory (<code>$XDG_DATA_HOME/powerhub/workspace/upload/</code> by default)." %}

{% set clipboard_main = "This is the Clipboard. It's for transfering small
code snippets, passwords, hashes, and so forth. You can also configure the
launcher to automatically execute contents of a particular clipboard entry." %}

{% set liststatic_main = "This is the Static Files List. PowerHub serves files
from PowerHub's data home directory (<code>$XDG_DATA_HOME/powerhub/static/</code>
by default). This page lists your static files and allows you to copy the URL
conveniently. Note that access to these files is not authenticated." %}
